@@ -12,6 +12,7 @@ _IS_QISKIT_AVAILABLE = True
 try:
     from qiskit.circuit.quantumcircuit import QuantumCircuit
     from qiskit.compiler.transpiler import transpile
+    from qiskit.circuit.library import U3Gate, UCGate
     import numpy as np
     import math
 except ImportError:
@@ -30,6 +31,23 @@ try:
 except ImportError:
     _IS_TENSORCIRCUIT_AVAILABLE = False
 
+
+def euler_angles_1q(m):
+    phase = (m[0][0] * m[1][1] - m[0][1] * m[1][0]) ** (-1.0/2.0)
+    U = [[phase * m[0][0], phase * m[0][1]], [phase * m[1][0], phase * m[1][1]]]
+
+    theta = 2 * math.atan2(abs(U[1][0]), abs(U[0][0]))
+
+    # Find phi and lambda
+    phiplambda = 2 * np.angle(U[1][1])
+    phimlambda = 2 * np.angle(U[1][0])
+
+    phi = (phiplambda + phimlambda) / 2.0
+    lamb = (phiplambda - phimlambda) / 2.0
+
+    return theta, phi, lamb
+
+
 class QrackCircuit:
     """Class that exposes the QCircuit class of Qrack
 
@@ -42,9 +60,13 @@ class QrackCircuit:
         cid(int): Qrack ID of this circuit
     """
 
-    def __init__(self, clone_cid = -1):
+    def __init__(self, is_collapse = True, is_near_clifford = False, clone_cid = -1, is_inverse=False, past_light_cone = []):
         if clone_cid < 0:
-            self.cid = Qrack.qrack_lib.init_qcircuit()
+            self.cid = Qrack.qrack_lib.init_qcircuit(is_collapse, is_near_clifford)
+        elif is_inverse:
+            self.cid = Qrack.qrack_lib.qcircuit_inverse(clone_cid)
+        elif len(past_light_cone) > 0:
+            self.cid = Qrack.qrack_lib.qcircuit_past_light_cone(clone_cid, len(past_light_cone), self._ulonglong_byref(past_light_cone))
         else:
             self.cid = Qrack.qrack_lib.init_qcircuit_clone(clone_cid)
 
@@ -63,11 +85,38 @@ class QrackCircuit:
         t = [(c.real, c.imag) for c in a]
         return self._double_byref([float(item) for sublist in t for item in sublist])
 
+    def clone(self):
+        """Make a new circuit that is an exact clone of this circuit
+
+        Raises:
+            RuntimeError: QrackCircuit C++ library raised an exception.
+        """
+        return QrackCircuit(clone_cid = self.cid, is_inverse = False)
+
+    def inverse(self):
+        """Make a new circuit that is the exact inverse of this circuit
+
+        Raises:
+            RuntimeError: QrackCircuit C++ library raised an exception.
+        """
+        return QrackCircuit(clone_cid = self.cid, is_inverse = True)
+
+    def past_light_cone(self, q):
+        """Make a new circuit with just this circuits' past light cone for certain qubits.
+
+        Args:
+            q: list of qubit indices to include at beginning of past light cone
+
+        Raises:
+            RuntimeError: QrackCircuit C++ library raised an exception.
+        """
+        return QrackCircuit(clone_cid = self.cid, is_inverse = False, past_light_cone = q)
+
     def get_qubit_count(self):
         """Get count of qubits in circuit
 
         Raises:
-            RuntimeError: QracQrackCircuitNeuron C++ library raised an exception.
+            RuntimeError: QrackCircuit C++ library raised an exception.
         """
         return Qrack.qrack_lib.get_qcircuit_qubit_count(self.cid)
 
@@ -168,6 +217,19 @@ class QrackCircuit:
 
         return out
 
+    def file_gate_count(filename):
+        """File gate count
+
+        Return the count of gates in a QrackCircuit file
+
+        Args:
+            filename: Name of file
+        """
+        tokens = []
+        with open(filename, 'r') as file:
+            tokens = file.read().split()
+        return int(tokens[1])
+
     def file_to_qiskit_circuit(filename):
         """Convert an output file to a Qiskit circuit
 
@@ -194,7 +256,7 @@ class QrackCircuit:
         i = 0
         num_qubits = int(tokens[i])
         i = i + 1
-        circ = QuantumCircuit(num_qubits, num_qubits)
+        circ = QuantumCircuit(num_qubits)
 
         num_gates = int(tokens[i])
         i = i + 1
@@ -254,13 +316,25 @@ class QrackCircuit:
                 payloads[key] = np.array(op)
 
             gate_list = []
-            for j in range(1 << control_count):
-                if j in payloads:
-                    gate_list.append(payloads[j])
-                else:
-                    gate_list.append(np.array([[1, 0],[0, 1]]))
-
-            circ.uc(gate_list, controls, target)
+            control_pow = 1 << control_count
+            pLen = len(payloads)
+            if (pLen == 1) or ((control_pow - pLen) > (1 << 15)):
+                for c, p in payloads.items():
+                    theta, phi, lam = euler_angles_1q(p)
+                    if control_count > 0:
+                        circ.append(
+                            U3Gate(theta, phi, lam).control(num_ctrl_qubits=control_count, ctrl_state=c),
+                            controls + [target]
+                        )
+                    else:
+                        circ.append(U3Gate(theta, phi, lam), [target])
+            else:
+                for j in range(control_pow):
+                    if j in payloads:
+                        gate_list.append(payloads[j])
+                    else:
+                        gate_list.append(np.array([[1, 0],[0, 1]]))
+                circ.append(UCGate(gate_list), controls + [target])
 
         return circ
 
