@@ -22,25 +22,41 @@ from sys import platform as _platform
 
 class QrackSystem:
     def __init__(self):
-        shared_lib_path = "/usr/local/lib/libqrack_pinvoke.so"
+        shared_lib_path = ""
         if os.environ.get('PYQRACK_SHARED_LIB_PATH') != None:
             shared_lib_path = os.environ.get('PYQRACK_SHARED_LIB_PATH')
-        elif _platform == "darwin":
-            shared_lib_path = "/usr/local/lib/libqrack_pinvoke.dylib"
         elif _platform == "win32":
-            shared_lib_path = "C:\\Program Files\\Qrack\\bin\\qrack_pinvoke.dll"
-        elif _platform != "linux" and _platform != "linux2":
-            print(
-                "No Qrack binary for your platform, attempting to use /usr/local/lib/libqrack_pinvoke.so"
-            )
-            print(
-                "You can choose the binary file to load with the environment variable: PYQRACK_SHARED_LIB_PATH"
-            )
+            shared_lib_path = os.path.dirname(__file__) + "/qrack_lib/qrack_pinvoke.dll"
+        elif _platform == "darwin":
+            shared_lib_path = os.path.dirname(__file__) + "/qrack_lib/libqrack_pinvoke.dylib"
+        else:
+            shared_lib_path = os.path.dirname(__file__) + "/qrack_lib/libqrack_pinvoke.so"
 
         try:
             self.qrack_lib = CDLL(shared_lib_path)
         except Exception as e:
-            print(e)
+            if _platform == "win32":
+                shared_lib_path = "C:/Program Files/libqrack*/lib/qrack_pinvoke.lib"
+            elif _platform == "darwin":
+                shared_lib_path = "/usr/local/lib/qrack/libqrack_pinvoke.dylib"
+            else:
+                shared_lib_path = "/usr/local/lib/qrack/libqrack_pinvoke.so"
+
+            try:
+                self.qrack_lib = CDLL(shared_lib_path)
+            except Exception as e:
+                if _platform == "win32":
+                    shared_lib_path = "C:/Program Files (x86)/libqrack*/lib/qrack_pinvoke.lib"
+                elif _platform == "darwin":
+                    shared_lib_path = "/usr/lib/qrack/libqrack_pinvoke.dylib"
+                else:
+                    shared_lib_path = "/usr/lib/qrack/libqrack_pinvoke.so"
+
+                try:
+                    self.qrack_lib = CDLL(shared_lib_path)
+                except Exception as e:
+                    print("IMPORTANT: Did you remember to install OpenCL, if your Qrack version was built with OpenCL?")
+                    raise e
 
         self.fppow = 5
         if "QRACK_FPPOW" in os.environ:
@@ -60,19 +76,22 @@ class QrackSystem:
         self.qrack_lib.Dump.restype = None
         self.qrack_lib.Dump.argtypes = [
             c_ulonglong,
-            CFUNCTYPE(c_ulonglong, c_double, c_double),
+            CFUNCTYPE(c_ulonglong, c_double, c_double)
         ]
 
         # These next two methods need to have c_double pointers, if PyQrack is built with fp64.
         self.qrack_lib.InKet.restype = None
         self.qrack_lib.OutKet.restype = None
+        self.qrack_lib.OutProbs.restype = None
 
-        if self.fppow == 5:
+        if self.fppow < 6:
             self.qrack_lib.InKet.argtypes = [c_ulonglong, POINTER(c_float)]
             self.qrack_lib.OutKet.argtypes = [c_ulonglong, POINTER(c_float)]
-        if self.fppow == 6:
+            self.qrack_lib.OutProbs.argtypes = [c_ulonglong, POINTER(c_float)]
+        else:
             self.qrack_lib.InKet.argtypes = [c_ulonglong, POINTER(c_double)]
             self.qrack_lib.OutKet.argtypes = [c_ulonglong, POINTER(c_double)]
+            self.qrack_lib.OutProbs.argtypes = [c_ulonglong, POINTER(c_double)]
 
         self.qrack_lib.init.restype = c_ulonglong
         self.qrack_lib.init.argtypes = []
@@ -98,20 +117,7 @@ class QrackSystem:
             c_bool,
             c_bool,
             c_bool,
-        ]
-
-        self.qrack_lib.init_count_type.restype = c_ulonglong
-        self.qrack_lib.init_count_type.argtypes = [
-            c_ulonglong,
-            c_bool,
-            c_bool,
-            c_bool,
-            c_bool,
-            c_bool,
-            c_bool,
-            c_bool,
-            c_bool,
-            c_bool,
+            c_bool
         ]
 
         self.qrack_lib.init_clone.restype = c_ulonglong
@@ -128,9 +134,28 @@ class QrackSystem:
 
         # pseudo-quantum
 
+        self.qrack_lib.ProbAll.restype = None
+        if self.fppow == 5:
+            self.qrack_lib.ProbAll.argtypes = [
+                c_ulonglong,
+                c_ulonglong,
+                POINTER(c_ulonglong),
+                POINTER(c_float)
+            ]
+        elif self.fppow == 6:
+            self.qrack_lib.ProbAll.argtypes = [
+                c_ulonglong,
+                c_ulonglong,
+                POINTER(c_ulonglong),
+                POINTER(c_double)
+            ]
+
         self.qrack_lib.Prob.restype = c_double
         self.qrack_lib.Prob.argtypes = [c_ulonglong, c_ulonglong]
 
+        self.qrack_lib.ProbRdm.restype = c_double
+        self.qrack_lib.ProbRdm.argtypes = [c_ulonglong, c_ulonglong]
+
         self.qrack_lib.PermutationProb.restype = c_double
         self.qrack_lib.PermutationProb.argtypes = [
             c_ulonglong,
@@ -139,11 +164,283 @@ class QrackSystem:
             POINTER(c_bool)
         ]
 
+        self.qrack_lib.PermutationProbRdm.restype = c_double
+        self.qrack_lib.PermutationProbRdm.argtypes = [
+            c_ulonglong,
+            c_ulonglong,
+            POINTER(c_ulonglong),
+            POINTER(c_bool),
+            c_bool
+        ]
+
         self.qrack_lib.PermutationExpectation.restype = c_double
         self.qrack_lib.PermutationExpectation.argtypes = [
+            c_ulonglong,
+            c_ulonglong,
+            POINTER(c_ulonglong)
+        ]
+
+        self.qrack_lib.PermutationExpectationRdm.restype = c_double
+        self.qrack_lib.PermutationExpectationRdm.argtypes = [
             c_ulonglong,
             c_ulonglong,
             POINTER(c_ulonglong),
+            c_bool
+        ]
+
+        self.qrack_lib.FactorizedExpectation.restype = c_double
+        self.qrack_lib.FactorizedExpectation.argtypes = [
+            c_ulonglong,
+            c_ulonglong,
+            POINTER(c_ulonglong),
+            c_ulonglong,
+            POINTER(c_ulonglong)
+        ]
+
+        self.qrack_lib.FactorizedExpectationRdm.restype = c_double
+        self.qrack_lib.FactorizedExpectationRdm.argtypes = [
+            c_ulonglong,
+            c_ulonglong,
+            POINTER(c_ulonglong),
+            c_ulonglong,
+            POINTER(c_ulonglong),
+            c_bool
+        ]
+
+        if self.fppow == 5:
+            self.qrack_lib.FactorizedExpectationFp.restype = c_double
+            self.qrack_lib.FactorizedExpectationFp.argtypes = [
+                c_ulonglong,
+                c_ulonglong,
+                POINTER(c_ulonglong),
+                POINTER(c_float)
+            ]
+            self.qrack_lib.FactorizedExpectationFpRdm.restype = c_double
+            self.qrack_lib.FactorizedExpectationFpRdm.argtypes = [
+                c_ulonglong,
+                c_ulonglong,
+                POINTER(c_ulonglong),
+                POINTER(c_float),
+                c_bool
+            ]
+            self.qrack_lib.UnitaryExpectation.restype = c_double
+            self.qrack_lib.UnitaryExpectation.argtypes = [
+                c_ulonglong,
+                c_ulonglong,
+                POINTER(c_ulonglong),
+                POINTER(c_float)
+            ]
+            self.qrack_lib.MatrixExpectation.restype = c_double
+            self.qrack_lib.MatrixExpectation.argtypes = [
+                c_ulonglong,
+                c_ulonglong,
+                POINTER(c_ulonglong),
+                POINTER(c_float)
+            ]
+            self.qrack_lib.UnitaryExpectationEigenVal.restype = c_double
+            self.qrack_lib.UnitaryExpectationEigenVal.argtypes = [
+                c_ulonglong,
+                c_ulonglong,
+                POINTER(c_ulonglong),
+                POINTER(c_float),
+                POINTER(c_float)
+            ]
+            self.qrack_lib.MatrixExpectationEigenVal.restype = c_double
+            self.qrack_lib.MatrixExpectationEigenVal.argtypes = [
+                c_ulonglong,
+                c_ulonglong,
+                POINTER(c_ulonglong),
+                POINTER(c_float),
+                POINTER(c_float)
+            ]
+        elif self.fppow == 6:
+            self.qrack_lib.FactorizedExpectationFp.restype = c_double
+            self.qrack_lib.FactorizedExpectationFp.argtypes = [
+                c_ulonglong,
+                c_ulonglong,
+                POINTER(c_ulonglong),
+                POINTER(c_double)
+            ]
+            self.qrack_lib.FactorizedExpectationFpRdm.restype = c_double
+            self.qrack_lib.FactorizedExpectationFpRdm.argtypes = [
+                c_ulonglong,
+                c_ulonglong,
+                POINTER(c_ulonglong),
+                POINTER(c_double),
+                c_bool
+            ]
+            self.qrack_lib.UnitaryExpectation.restype = c_double
+            self.qrack_lib.UnitaryExpectation.argtypes = [
+                c_ulonglong,
+                c_ulonglong,
+                POINTER(c_ulonglong),
+                POINTER(c_double)
+            ]
+            self.qrack_lib.MatrixExpectation.restype = c_double
+            self.qrack_lib.MatrixExpectation.argtypes = [
+                c_ulonglong,
+                c_ulonglong,
+                POINTER(c_ulonglong),
+                POINTER(c_double)
+            ]
+            self.qrack_lib.UnitaryExpectationEigenVal.restype = c_double
+            self.qrack_lib.UnitaryExpectationEigenVal.argtypes = [
+                c_ulonglong,
+                c_ulonglong,
+                POINTER(c_ulonglong),
+                POINTER(c_double),
+                POINTER(c_double)
+            ]
+            self.qrack_lib.MatrixExpectationEigenVal.restype = c_double
+            self.qrack_lib.MatrixExpectationEigenVal.argtypes = [
+                c_ulonglong,
+                c_ulonglong,
+                POINTER(c_ulonglong),
+                POINTER(c_double),
+                POINTER(c_double)
+            ]
+
+        self.qrack_lib.PauliExpectation.restype = c_double
+        self.qrack_lib.PauliExpectation.argtypes = [
+            c_ulonglong,
+            c_ulonglong,
+            POINTER(c_ulonglong),
+            POINTER(c_ulonglong)
+        ]
+
+        self.qrack_lib.Variance.restype = c_double
+        self.qrack_lib.Variance.argtypes = [
+            c_ulonglong,
+            c_ulonglong,
+            POINTER(c_ulonglong)
+        ]
+
+        self.qrack_lib.VarianceRdm.restype = c_double
+        self.qrack_lib.VarianceRdm.argtypes = [
+            c_ulonglong,
+            c_ulonglong,
+            POINTER(c_ulonglong),
+            c_bool
+        ]
+
+        self.qrack_lib.FactorizedVariance.restype = c_double
+        self.qrack_lib.FactorizedVariance.argtypes = [
+            c_ulonglong,
+            c_ulonglong,
+            POINTER(c_ulonglong),
+            c_ulonglong,
+            POINTER(c_ulonglong)
+        ]
+
+        self.qrack_lib.FactorizedVarianceRdm.restype = c_double
+        self.qrack_lib.FactorizedVarianceRdm.argtypes = [
+            c_ulonglong,
+            c_ulonglong,
+            POINTER(c_ulonglong),
+            c_ulonglong,
+            POINTER(c_ulonglong),
+            c_bool
+        ]
+
+        if self.fppow == 5:
+            self.qrack_lib.FactorizedVarianceFp.restype = c_double
+            self.qrack_lib.FactorizedVarianceFp.argtypes = [
+                c_ulonglong,
+                c_ulonglong,
+                POINTER(c_ulonglong),
+                POINTER(c_float)
+            ]
+            self.qrack_lib.FactorizedVarianceFpRdm.restype = c_double
+            self.qrack_lib.FactorizedVarianceFpRdm.argtypes = [
+                c_ulonglong,
+                c_ulonglong,
+                POINTER(c_ulonglong),
+                POINTER(c_float),
+                c_bool
+            ]
+            self.qrack_lib.UnitaryVariance.restype = c_double
+            self.qrack_lib.UnitaryVariance.argtypes = [
+                c_ulonglong,
+                c_ulonglong,
+                POINTER(c_ulonglong),
+                POINTER(c_float)
+            ]
+            self.qrack_lib.MatrixVariance.restype = c_double
+            self.qrack_lib.MatrixVariance.argtypes = [
+                c_ulonglong,
+                c_ulonglong,
+                POINTER(c_ulonglong),
+                POINTER(c_float)
+            ]
+            self.qrack_lib.UnitaryVarianceEigenVal.restype = c_double
+            self.qrack_lib.UnitaryVarianceEigenVal.argtypes = [
+                c_ulonglong,
+                c_ulonglong,
+                POINTER(c_ulonglong),
+                POINTER(c_float),
+                POINTER(c_float)
+            ]
+            self.qrack_lib.MatrixVarianceEigenVal.restype = c_double
+            self.qrack_lib.MatrixVarianceEigenVal.argtypes = [
+                c_ulonglong,
+                c_ulonglong,
+                POINTER(c_ulonglong),
+                POINTER(c_float),
+                POINTER(c_float)
+            ]
+        elif self.fppow == 6:
+            self.qrack_lib.FactorizedVarianceFp.restype = c_double
+            self.qrack_lib.FactorizedVarianceFp.argtypes = [
+                c_ulonglong,
+                c_ulonglong,
+                POINTER(c_ulonglong),
+                POINTER(c_double)
+            ]
+            self.qrack_lib.FactorizedVarianceFpRdm.restype = c_double
+            self.qrack_lib.FactorizedVarianceFpRdm.argtypes = [
+                c_ulonglong,
+                c_ulonglong,
+                POINTER(c_ulonglong),
+                POINTER(c_double),
+                c_bool
+            ]
+            self.qrack_lib.UnitaryVariance.restype = c_double
+            self.qrack_lib.UnitaryVariance.argtypes = [
+                c_ulonglong,
+                c_ulonglong,
+                POINTER(c_ulonglong),
+                POINTER(c_double)
+            ]
+            self.qrack_lib.MatrixVariance.restype = c_double
+            self.qrack_lib.MatrixVariance.argtypes = [
+                c_ulonglong,
+                c_ulonglong,
+                POINTER(c_ulonglong),
+                POINTER(c_double)
+            ]
+            self.qrack_lib.UnitaryVarianceEigenVal.restype = c_double
+            self.qrack_lib.UnitaryVarianceEigenVal.argtypes = [
+                c_ulonglong,
+                c_ulonglong,
+                POINTER(c_ulonglong),
+                POINTER(c_double),
+                POINTER(c_double)
+            ]
+            self.qrack_lib.MatrixVarianceEigenVal.restype = c_double
+            self.qrack_lib.MatrixVarianceEigenVal.argtypes = [
+                c_ulonglong,
+                c_ulonglong,
+                POINTER(c_ulonglong),
+                POINTER(c_double),
+                POINTER(c_double)
+            ]
+
+        self.qrack_lib.PauliVariance.restype = c_double
+        self.qrack_lib.PauliVariance.argtypes = [
+            c_ulonglong,
+            c_ulonglong,
+            POINTER(c_ulonglong),
+            POINTER(c_ulonglong)
         ]
 
         self.qrack_lib.JointEnsembleProbability.restype = c_double
@@ -151,7 +448,7 @@ class QrackSystem:
             c_ulonglong,
             c_ulonglong,
             POINTER(c_int),
-            c_ulonglong,
+            c_ulonglong
         ]
 
         self.qrack_lib.PhaseParity.restype = None
@@ -159,7 +456,15 @@ class QrackSystem:
             c_ulonglong,
             c_double,
             c_ulonglong,
-            POINTER(c_ulonglong),
+            POINTER(c_ulonglong)
+        ]
+
+        self.qrack_lib.PhaseRootN.restype = None
+        self.qrack_lib.PhaseRootN.argtypes = [
+            c_ulonglong,
+            c_ulonglong,
+            c_ulonglong,
+            POINTER(c_ulonglong)
         ]
 
         self.qrack_lib.ResetAll.restype = None
@@ -208,7 +513,7 @@ class QrackSystem:
             c_ulonglong,
             c_double,
             c_double,
-            c_double,
+            c_double
         ]
 
         self.qrack_lib.Mtrx.restype = None
@@ -221,7 +526,7 @@ class QrackSystem:
             c_ulonglong,
             c_ulonglong,
             POINTER(c_ulonglong),
-            c_ulonglong,
+            c_ulonglong
         ]
 
         self.qrack_lib.MCY.restype = None
@@ -229,7 +534,7 @@ class QrackSystem:
             c_ulonglong,
             c_ulonglong,
             POINTER(c_ulonglong),
-            c_ulonglong,
+            c_ulonglong
         ]
 
         self.qrack_lib.MCZ.restype = None
@@ -237,7 +542,7 @@ class QrackSystem:
             c_ulonglong,
             c_ulonglong,
             POINTER(c_ulonglong),
-            c_ulonglong,
+            c_ulonglong
         ]
 
         self.qrack_lib.MCH.restype = None
@@ -245,7 +550,7 @@ class QrackSystem:
             c_ulonglong,
             c_ulonglong,
             POINTER(c_ulonglong),
-            c_ulonglong,
+            c_ulonglong
         ]
 
         self.qrack_lib.MCS.restype = None
@@ -253,7 +558,7 @@ class QrackSystem:
             c_ulonglong,
             c_ulonglong,
             POINTER(c_ulonglong),
-            c_ulonglong,
+            c_ulonglong
         ]
 
         self.qrack_lib.MCT.restype = None
@@ -261,7 +566,7 @@ class QrackSystem:
             c_ulonglong,
             c_ulonglong,
             POINTER(c_ulonglong),
-            c_ulonglong,
+            c_ulonglong
         ]
 
         self.qrack_lib.MCAdjS.restype = None
@@ -269,7 +574,7 @@ class QrackSystem:
             c_ulonglong,
             c_ulonglong,
             POINTER(c_ulonglong),
-            c_ulonglong,
+            c_ulonglong
         ]
 
         self.qrack_lib.MCAdjT.restype = None
@@ -277,7 +582,7 @@ class QrackSystem:
             c_ulonglong,
             c_ulonglong,
             POINTER(c_ulonglong),
-            c_ulonglong,
+            c_ulonglong
         ]
 
         self.qrack_lib.MCU.restype = None
@@ -288,7 +593,7 @@ class QrackSystem:
             c_ulonglong,
             c_double,
             c_double,
-            c_double,
+            c_double
         ]
 
         self.qrack_lib.MCMtrx.restype = None
@@ -297,7 +602,7 @@ class QrackSystem:
             c_ulonglong,
             POINTER(c_ulonglong),
             POINTER(c_double),
-            c_ulonglong,
+            c_ulonglong
         ]
 
         # multi-anti-controlled single-qubit gates
@@ -307,7 +612,7 @@ class QrackSystem:
             c_ulonglong,
             c_ulonglong,
             POINTER(c_ulonglong),
-            c_ulonglong,
+            c_ulonglong
         ]
 
         self.qrack_lib.MACY.restype = None
@@ -315,7 +620,7 @@ class QrackSystem:
             c_ulonglong,
             c_ulonglong,
             POINTER(c_ulonglong),
-            c_ulonglong,
+            c_ulonglong
         ]
 
         self.qrack_lib.MACZ.restype = None
@@ -323,7 +628,7 @@ class QrackSystem:
             c_ulonglong,
             c_ulonglong,
             POINTER(c_ulonglong),
-            c_ulonglong,
+            c_ulonglong
         ]
 
         self.qrack_lib.MACH.restype = None
@@ -331,7 +636,7 @@ class QrackSystem:
             c_ulonglong,
             c_ulonglong,
             POINTER(c_ulonglong),
-            c_ulonglong,
+            c_ulonglong
         ]
 
         self.qrack_lib.MACS.restype = None
@@ -339,7 +644,7 @@ class QrackSystem:
             c_ulonglong,
             c_ulonglong,
             POINTER(c_ulonglong),
-            c_ulonglong,
+            c_ulonglong
         ]
 
         self.qrack_lib.MACT.restype = None
@@ -347,7 +652,7 @@ class QrackSystem:
             c_ulonglong,
             c_ulonglong,
             POINTER(c_ulonglong),
-            c_ulonglong,
+            c_ulonglong
         ]
 
         self.qrack_lib.MACAdjS.restype = None
@@ -355,7 +660,7 @@ class QrackSystem:
             c_ulonglong,
             c_ulonglong,
             POINTER(c_ulonglong),
-            c_ulonglong,
+            c_ulonglong
         ]
 
         self.qrack_lib.MACAdjT.restype = None
@@ -363,7 +668,7 @@ class QrackSystem:
             c_ulonglong,
             c_ulonglong,
             POINTER(c_ulonglong),
-            c_ulonglong,
+            c_ulonglong
         ]
 
         self.qrack_lib.MACU.restype = None
@@ -374,7 +679,7 @@ class QrackSystem:
             c_ulonglong,
             c_double,
             c_double,
-            c_double,
+            c_double
         ]
 
         self.qrack_lib.MACMtrx.restype = None
@@ -383,7 +688,7 @@ class QrackSystem:
             c_ulonglong,
             POINTER(c_ulonglong),
             POINTER(c_double),
-            c_ulonglong,
+            c_ulonglong
         ]
 
         self.qrack_lib.UCMtrx.restype = None
@@ -393,7 +698,7 @@ class QrackSystem:
             POINTER(c_ulonglong),
             POINTER(c_double),
             c_ulonglong,
-            c_ulonglong,
+            c_ulonglong
         ]
 
         self.qrack_lib.Multiplex1Mtrx.restype = None
@@ -402,7 +707,7 @@ class QrackSystem:
             c_ulonglong,
             POINTER(c_ulonglong),
             c_ulonglong,
-            POINTER(c_double),
+            POINTER(c_double)
         ]
 
         # coalesced single qubit gates
@@ -428,7 +733,7 @@ class QrackSystem:
             c_double,
             c_ulonglong,
             POINTER(c_ulonglong),
-            c_ulonglong,
+            c_ulonglong
         ]
 
         # exponential of Pauli operators
@@ -439,7 +744,7 @@ class QrackSystem:
             c_ulonglong,
             POINTER(c_int),
             c_double,
-            POINTER(c_ulonglong),
+            POINTER(c_ulonglong)
         ]
 
         self.qrack_lib.MCExp.restype = None
@@ -450,7 +755,7 @@ class QrackSystem:
             c_double,
             c_ulonglong,
             POINTER(c_ulonglong),
-            POINTER(c_ulonglong),
+            POINTER(c_ulonglong)
         ]
 
         # measurements
@@ -469,7 +774,7 @@ class QrackSystem:
             c_ulonglong,
             c_ulonglong,
             POINTER(c_int),
-            POINTER(c_ulonglong),
+            POINTER(c_ulonglong)
         ]
 
         self.qrack_lib.MeasureShots.restype = None
@@ -478,7 +783,7 @@ class QrackSystem:
             c_ulonglong,
             POINTER(c_ulonglong),
             c_ulonglong,
-            POINTER(c_ulonglong),
+            POINTER(c_ulonglong)
         ]
 
         # swap
@@ -498,7 +803,7 @@ class QrackSystem:
             c_double,
             c_double,
             c_ulonglong,
-            c_ulonglong,
+            c_ulonglong
         ]
 
         self.qrack_lib.CSWAP.restype = None
@@ -507,7 +812,7 @@ class QrackSystem:
             c_ulonglong,
             POINTER(c_ulonglong),
             c_ulonglong,
-            c_ulonglong,
+            c_ulonglong
         ]
 
         self.qrack_lib.ACSWAP.restype = None
@@ -516,7 +821,7 @@ class QrackSystem:
             c_ulonglong,
             POINTER(c_ulonglong),
             c_ulonglong,
-            c_ulonglong,
+            c_ulonglong
         ]
 
         # Schmidt decomposition
@@ -525,21 +830,21 @@ class QrackSystem:
         self.qrack_lib.Compose.argtypes = [
             c_ulonglong,
             c_ulonglong,
-            POINTER(c_ulonglong),
+            POINTER(c_ulonglong)
         ]
 
         self.qrack_lib.Decompose.restype = c_ulonglong
         self.qrack_lib.Decompose.argtypes = [
             c_ulonglong,
             c_ulonglong,
-            POINTER(c_ulonglong),
+            POINTER(c_ulonglong)
         ]
 
         self.qrack_lib.Dispose.restype = None
         self.qrack_lib.Dispose.argtypes = [
             c_ulonglong,
             c_ulonglong,
-            POINTER(c_ulonglong),
+            POINTER(c_ulonglong)
         ]
 
         # (quasi-)Boolean gates
@@ -549,7 +854,7 @@ class QrackSystem:
             c_ulonglong,
             c_ulonglong,
             c_ulonglong,
-            c_ulonglong,
+            c_ulonglong
         ]
 
         self.qrack_lib.OR.restype = None
@@ -557,7 +862,7 @@ class QrackSystem:
             c_ulonglong,
             c_ulonglong,
             c_ulonglong,
-            c_ulonglong,
+            c_ulonglong
         ]
 
         self.qrack_lib.XOR.restype = None
@@ -565,7 +870,7 @@ class QrackSystem:
             c_ulonglong,
             c_ulonglong,
             c_ulonglong,
-            c_ulonglong,
+            c_ulonglong
         ]
 
         self.qrack_lib.NAND.restype = None
@@ -573,7 +878,7 @@ class QrackSystem:
             c_ulonglong,
             c_ulonglong,
             c_ulonglong,
-            c_ulonglong,
+            c_ulonglong
         ]
 
         self.qrack_lib.NOR.restype = None
@@ -581,7 +886,7 @@ class QrackSystem:
             c_ulonglong,
             c_ulonglong,
             c_ulonglong,
-            c_ulonglong,
+            c_ulonglong
         ]
 
         self.qrack_lib.XNOR.restype = None
@@ -589,7 +894,7 @@ class QrackSystem:
             c_ulonglong,
             c_ulonglong,
             c_ulonglong,
-            c_ulonglong,
+            c_ulonglong
         ]
 
         # half classical (quasi-)Boolean gates
@@ -628,7 +933,7 @@ class QrackSystem:
             c_ulonglong,
             POINTER(c_ulonglong),
             c_ulonglong,
-            POINTER(c_ulonglong),
+            POINTER(c_ulonglong)
         ]
 
         self.qrack_lib.SUB.restype = None
@@ -637,7 +942,7 @@ class QrackSystem:
             c_ulonglong,
             POINTER(c_ulonglong),
             c_ulonglong,
-            POINTER(c_ulonglong),
+            POINTER(c_ulonglong)
         ]
 
         self.qrack_lib.ADDS.restype = None
@@ -647,7 +952,7 @@ class QrackSystem:
             POINTER(c_ulonglong),
             c_ulonglong,
             c_ulonglong,
-            POINTER(c_ulonglong),
+            POINTER(c_ulonglong)
         ]
 
         self.qrack_lib.SUBS.restype = None
@@ -657,7 +962,7 @@ class QrackSystem:
             POINTER(c_ulonglong),
             c_ulonglong,
             c_ulonglong,
-            POINTER(c_ulonglong),
+            POINTER(c_ulonglong)
         ]
 
         self.qrack_lib.MUL.restype = None
@@ -667,7 +972,7 @@ class QrackSystem:
             POINTER(c_ulonglong),
             c_ulonglong,
             POINTER(c_ulonglong),
-            POINTER(c_ulonglong),
+            POINTER(c_ulonglong)
         ]
 
         self.qrack_lib.DIV.restype = None
@@ -677,7 +982,7 @@ class QrackSystem:
             POINTER(c_ulonglong),
             c_ulonglong,
             POINTER(c_ulonglong),
-            POINTER(c_ulonglong),
+            POINTER(c_ulonglong)
         ]
 
         self.qrack_lib.MULN.restype = None
@@ -688,7 +993,7 @@ class QrackSystem:
             POINTER(c_ulonglong),
             c_ulonglong,
             POINTER(c_ulonglong),
-            POINTER(c_ulonglong),
+            POINTER(c_ulonglong)
         ]
 
         self.qrack_lib.DIVN.restype = None
@@ -699,7 +1004,7 @@ class QrackSystem:
             POINTER(c_ulonglong),
             c_ulonglong,
             POINTER(c_ulonglong),
-            POINTER(c_ulonglong),
+            POINTER(c_ulonglong)
         ]
 
         self.qrack_lib.POWN.restype = None
@@ -710,7 +1015,7 @@ class QrackSystem:
             POINTER(c_ulonglong),
             c_ulonglong,
             POINTER(c_ulonglong),
-            POINTER(c_ulonglong),
+            POINTER(c_ulonglong)
         ]
 
         self.qrack_lib.MCADD.restype = None
@@ -721,7 +1026,7 @@ class QrackSystem:
             c_ulonglong,
             POINTER(c_ulonglong),
             c_ulonglong,
-            POINTER(c_ulonglong),
+            POINTER(c_ulonglong)
         ]
 
         self.qrack_lib.MCSUB.restype = None
@@ -732,7 +1037,7 @@ class QrackSystem:
             c_ulonglong,
             POINTER(c_ulonglong),
             c_ulonglong,
-            POINTER(c_ulonglong),
+            POINTER(c_ulonglong)
         ]
 
         self.qrack_lib.MCMUL.restype = None
@@ -743,7 +1048,7 @@ class QrackSystem:
             c_ulonglong,
             POINTER(c_ulonglong),
             c_ulonglong,
-            POINTER(c_ulonglong),
+            POINTER(c_ulonglong)
         ]
 
         self.qrack_lib.MCDIV.restype = None
@@ -754,7 +1059,7 @@ class QrackSystem:
             c_ulonglong,
             POINTER(c_ulonglong),
             c_ulonglong,
-            POINTER(c_ulonglong),
+            POINTER(c_ulonglong)
         ]
 
         self.qrack_lib.MCMULN.restype = None
@@ -767,7 +1072,7 @@ class QrackSystem:
             POINTER(c_ulonglong),
             c_ulonglong,
             POINTER(c_ulonglong),
-            POINTER(c_ulonglong),
+            POINTER(c_ulonglong)
         ]
 
         self.qrack_lib.MCDIVN.restype = None
@@ -780,7 +1085,7 @@ class QrackSystem:
             POINTER(c_ulonglong),
             c_ulonglong,
             POINTER(c_ulonglong),
-            POINTER(c_ulonglong),
+            POINTER(c_ulonglong)
         ]
 
         self.qrack_lib.MCPOWN.restype = None
@@ -793,7 +1098,7 @@ class QrackSystem:
             POINTER(c_ulonglong),
             c_ulonglong,
             POINTER(c_ulonglong),
-            POINTER(c_ulonglong),
+            POINTER(c_ulonglong)
         ]
 
         self.qrack_lib.LDA.restype = None
@@ -803,7 +1108,7 @@ class QrackSystem:
             POINTER(c_ulonglong),
             c_ulonglong,
             POINTER(c_ulonglong),
-            POINTER(c_ubyte),
+            POINTER(c_ubyte)
         ]
 
         self.qrack_lib.ADC.restype = None
@@ -814,7 +1119,7 @@ class QrackSystem:
             POINTER(c_ulonglong),
             c_ulonglong,
             POINTER(c_ulonglong),
-            POINTER(c_ubyte),
+            POINTER(c_ubyte)
         ]
 
         self.qrack_lib.SBC.restype = None
@@ -825,7 +1130,7 @@ class QrackSystem:
             POINTER(c_ulonglong),
             c_ulonglong,
             POINTER(c_ulonglong),
-            POINTER(c_ubyte),
+            POINTER(c_ubyte)
         ]
 
         self.qrack_lib.Hash.restype = None
@@ -833,7 +1138,7 @@ class QrackSystem:
             c_ulonglong,
             c_ulonglong,
             POINTER(c_ulonglong),
-            POINTER(c_ubyte),
+            POINTER(c_ubyte)
         ]
 
         # miscellaneous
@@ -849,7 +1154,14 @@ class QrackSystem:
             c_ulonglong,
             c_ulonglong,
             POINTER(c_ulonglong),
-            c_double,
+            c_double
+        ]
+
+        self.qrack_lib.Separate.restype = None
+        self.qrack_lib.Separate.argtypes = [
+            c_ulonglong,
+            c_ulonglong,
+            POINTER(c_ulonglong)
         ]
 
         self.qrack_lib.GetUnitaryFidelity.restype = c_double
@@ -861,14 +1173,20 @@ class QrackSystem:
         self.qrack_lib.SetSdrp.restype = None
         self.qrack_lib.SetSdrp.argtypes = [c_ulonglong, c_double]
 
-        self.qrack_lib.SetReactiveSeparate.restype = c_bool
+        self.qrack_lib.SetNcrp.restype = None
+        self.qrack_lib.SetNcrp.argtypes = [c_ulonglong, c_double]
+
+        self.qrack_lib.SetReactiveSeparate.restype = None
         self.qrack_lib.SetReactiveSeparate.argtypes = [c_ulonglong, c_bool]
 
-        self.qrack_lib.SetTInjection.restype = c_bool
+        self.qrack_lib.SetTInjection.restype = None
         self.qrack_lib.SetTInjection.argtypes = [c_ulonglong, c_bool]
 
-        self.qrack_lib.SetStabilizerHardwareEncoded.restype = c_bool
-        self.qrack_lib.SetStabilizerHardwareEncoded.argtypes = [c_ulonglong, c_bool]
+        self.qrack_lib.SetNoiseParameter.restype = None
+        self.qrack_lib.SetNoiseParameter.argtypes = [c_ulonglong, c_double]
+
+        self.qrack_lib.Normalize.restype = None
+        self.qrack_lib.Normalize.argtypes = [c_ulonglong]
 
         self.qrack_lib.qstabilizer_out_to_file.restype = None
         self.qrack_lib.qstabilizer_out_to_file.argtypes = [c_ulonglong, c_char_p]
@@ -891,7 +1209,7 @@ class QrackSystem:
         if self.fppow == 5:
             self.qrack_lib.set_qneuron_angles.argtypes = [c_ulonglong, POINTER(c_float)]
             self.qrack_lib.get_qneuron_angles.argtypes = [c_ulonglong, POINTER(c_float)]
-        if self.fppow == 6:
+        elif self.fppow == 6:
             self.qrack_lib.set_qneuron_angles.argtypes = [c_ulonglong, POINTER(c_double)]
             self.qrack_lib.get_qneuron_angles.argtypes = [c_ulonglong, POINTER(c_double)]
 
@@ -923,11 +1241,17 @@ class QrackSystem:
         self.qrack_lib.qneuron_learn_permutation.argtypes = [c_ulonglong, c_double, c_bool, c_bool]
 
         self.qrack_lib.init_qcircuit.restype = c_ulonglong
-        self.qrack_lib.init_qcircuit.argtypes = []
+        self.qrack_lib.init_qcircuit.argtypes = [c_bool, c_bool]
 
         self.qrack_lib.init_qcircuit_clone.restype = c_ulonglong
         self.qrack_lib.init_qcircuit_clone.argtypes = [c_ulonglong]
 
+        self.qrack_lib.qcircuit_inverse.restype = c_ulonglong
+        self.qrack_lib.qcircuit_inverse.argtypes = [c_ulonglong]
+
+        self.qrack_lib.qcircuit_past_light_cone.restype = c_ulonglong
+        self.qrack_lib.qcircuit_past_light_cone.argtypes = [c_ulonglong, c_ulonglong, POINTER(c_ulonglong)]
+
         self.qrack_lib.destroy_qcircuit.restype = None
         self.qrack_lib.destroy_qcircuit.argtypes = [c_ulonglong]
 
